////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2017 Josef Spidlen, Ph.D., FlowJo, LLC
//
// License
// The software is distributed under the terms of the 
// Artistic License 2.0
// http://www.r-project.org/Licenses/Artistic-2.0
// 
// Disclaimer
// This software and documentation come with no warranties of any kind.
// This software is provided "as is" and any express or implied 
// warranties, including, but not limited to, the implied warranties of
// merchantability and fitness for a particular purpose are disclaimed.
// In no event shall the  copyright holder be liable for any direct, 
// indirect, incidental, special, exemplary, or consequential damages
// (including but not limited to, procurement of substitute goods or 
// services; loss of use, data or profits; or business interruption)
// however caused and on any theory of liability, whether in contract,
// strict liability, or tort arising in any way out of the use of this 
// software.    
//////////////////////////////////////////////////////////////////////////////

package com.flowjo.plugin.flowai.utils;

import java.io.File;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import com.flowjo.plugin.flowai.FlowAI;

public abstract class FilenameUtils {

	public static String fixFileNamePart(String fileName) {
		return fileName.
			replaceAll(" ", "_").replaceAll("/", "_").replaceAll("\\\\", "_").
	        replaceAll("\\:", "_").replaceAll("\\*", "_").replaceAll("\\?", "_").
	        replaceAll("\"", "_").replaceAll("\\|", "_");
	}
	
	public static String createOutputFolderPath(String prefix, String sampleName, List<String> parameterNames, Map<String, String> options) {
		
		// This method needs to capture the sampleName as well as other input
		// of the flowClean algorithm so that results caching works properly
        String outFolder = System.getProperty("java.io.tmpdir");
        String fileSep = System.getProperty("file.separator");
        if(!outFolder.endsWith("/") && !outFolder.endsWith("\\")) 
        	outFolder = (new StringBuilder()).append(outFolder).append(fileSep).toString();
        
        Collections.sort(parameterNames);
        String parametersString = "";
        boolean first = true;
        for (String s : parameterNames)
        {
        	if(first) parametersString += s;
        	else parametersString += "_" + s;
        	first = false;
        }
        parametersString = fixFileNamePart(parametersString);
        
        outFolder = (new StringBuilder()).append(outFolder).
        	append(prefix).append(fileSep).
        	append(fixFileNamePart(sampleName)).append(fileSep).
        	append(options.get(FlowAI.sOptionNameRemoveFrom)).append("_").
        	append(options.get(FlowAI.sOptionNameSecondFractionFR)).append("_").
        	append(options.get(FlowAI.sOptionNameAlphaFR)).append("_").
        	append(options.get(FlowAI.sOptionNameMaxCptFs)).append("_").
        	append(options.get(FlowAI.sOptionNamePenValueFs)).append("_").
        	append(options.get(FlowAI.sOptionNameDecompFR)).append("_").
        	append(options.get(FlowAI.sOptionNameOutlierFS)).append("_").
        	append(options.get(FlowAI.sOptionNameDeleteWD)).append("_").
        	append(options.get(FlowAI.sOptionNameSideFM)).append("_").
        	append(parametersString).append(fileSep).
        	toString().replaceAll(" ", "_");
        
        new File(outFolder).mkdirs();
        return outFolder;
	}
}
